import os

_APP = os.path.abspath(os.path.dirname(__file__))
dir_data = os.path.join(_APP, "data")

data_atoms = {}
data_colors = {}

with open(os.path.join(dir_data, "atom_data")) as ifile:
    line = ifile.readline()
    while line != "":
        stripped = line.strip()
        if stripped.startswith("#"):
            pass
        elif stripped.startswith("C"):
            tmp = stripped.split()
            try:
                data_colors[tmp[1].upper()] = (float(tmp[2]), float(tmp[3]), float(tmp[4]), int(tmp[5]), int(tmp[6]))
            except (NameError, ValueError, KeyError):
                raise RuntimeError("DEFINITION FILE FORMAT ERROR (Colors)")
        elif stripped.startswith("A"):
            tmp = stripped.split()
            try:
                data_atoms[tmp[1].upper()] = (float(tmp[2]), tmp[3].upper())
            except (NameError, ValueError, KeyError):
                raise RuntimeError("DEFINITION FILE FORMAT ERROR (Covalent Radii)")
        line = ifile.readline()


def map_radii(a):
    for symbol in a:
        symbol = symbol.upper()
        if symbol in data_atoms:
            yield data_atoms[symbol][0]
        else:
            yield 1.5


def map_colors(a):
    for symbol in a:
        symbol = symbol.upper()
        if symbol in data_atoms:
            yield data_colors[data_atoms[symbol][1]]
        else:
            yield 0.9, 0.9, 0.9, 0, 0



import curses
from math import ceil
from time import sleep

from asciimol.app.config import conf
from asciimol.app.renderer import Renderer


class AsciiMol:
    def __init__(self):
        self.stdscr = None
        self.renderer = None
        self.sig_changed = True
        self.frames = 0
        self.timeout = 0

    def redraw(self):
        self.stdscr.clear()
        self.draw_characters(self.renderer.content)
        self.draw_navbar()
        self.stdscr.refresh()
        self.sig_changed = False

    def draw_characters(self, content: list):
        for i in range(curses.LINES - 2):
            for j in range(curses.COLS):
                char, col = content[i][j].split(",")
                self.stdscr.addstr(i, j, char, int(col))

    def draw_navbar(self):
        x, y, z = self.renderer.rotcounter
        ztoggle_str = "Z" if self.renderer.ztoggle else "Y"
        navbar_string = "[Q]uit [R]eset "
        navbar_string += "[B]onds %s " % ("on " if self.renderer.btoggle else "off")
        navbar_string += "[+-] Zoom (%- 3.3f) " % self.renderer.zoom
        navbar_string += "[↔↕] Rotate (%-3.f, %-3.f, %-3.f) " % (x, y, z)
        navbar_string += "[Z] ↔ Y/Z rotation (%s) " % ztoggle_str
        navbar_string += "[WSAD] Navigate "
        navbar_string += "[C] Center "
        navbar_string += "[T] Principle Axes "
        navbar_string += "[F1-3] Auto-Rotate"
        try:
            self.stdscr.addstr(navbar_string)
        except curses.error:
            if curses.LINES > 1 and curses.COLS > 3:
                self.stdscr.addstr(curses.LINES - 1, curses.COLS - 4, "...")

    def handle_keypresses(self, keys):
        self.sig_changed = False
        if len(keys) > 0:
            if 87 in keys or 119 in keys:  # W
                self.sig_changed = self.renderer.navigate(dy=-ceil(self.renderer.zoom))
            if 83 in keys or 115 in keys:  # S
                self.sig_changed = self.renderer.navigate(dy=ceil(self.renderer.zoom))
            if 65 in keys or 97 in keys:  # A
                self.sig_changed = self.renderer.navigate(dx=-ceil(self.renderer.zoom))
            if 68 in keys or 100 in keys:  # D
                self.sig_changed = self.renderer.navigate(dx=ceil(self.renderer.zoom))
            if 67 in keys or 99 in keys:  # C
                self.sig_changed = self.renderer.center()
            if 84 in keys or 116 in keys:  # T
                self.sig_changed = self.renderer.prinicple_axes()
            if curses.KEY_F1 in keys:
                self.renderer.toggle_auto_rotate(x=True)
            if curses.KEY_F2 in keys:
                self.renderer.toggle_auto_rotate(y=True)
            if curses.KEY_F3 in keys:
                self.renderer.toggle_auto_rotate(z=True)
            if curses.KEY_DOWN in keys:
                self.sig_changed = self.renderer.rotate(x=1)
            if curses.KEY_UP in keys:
                self.sig_changed = self.renderer.rotate(x=-1)
            if curses.KEY_LEFT in keys:
                self.sig_changed = self.renderer.rotate(z=1) if self.renderer.ztoggle else self.renderer.rotate(y=1)
            if curses.KEY_RIGHT in keys:
                self.sig_changed = self.renderer.rotate(z=-1) if self.renderer.ztoggle else self.renderer.rotate(y=-1)
            if 43 in keys:  # +
                self.sig_changed = self.renderer.modify_zoom(0.1)
            if 45 in keys:  # -
                self.sig_changed = self.renderer.modify_zoom(-0.1)
            if 82 in keys or 114 in keys:  # R
                self.sig_changed = self.renderer.reset_view()
            if 66 in keys or 98 in keys:  # B
                self.renderer.btoggle = not self.renderer.btoggle
                self.sig_changed = True
            if 90 in keys or 122 in keys:  # Z
                self.renderer.ztoggle = not self.renderer.ztoggle
                self.sig_changed = True
            if 81 in keys or 113 in keys:  # Q
                return False
        return True

    def on_update(self):
        keys = []
        key = self.stdscr.getch()
        while key is not curses.ERR:
            keys.append(key)
            key = self.stdscr.getch()
        running = self.handle_keypresses(keys)
        # Limit resizing to once per second, this is easier on curses.
        if self.frames == 59:
            if curses.is_term_resized(self.renderer.height, self.renderer.width):
                curses.update_lines_cols()
                self.renderer.resize(curses.LINES, curses.COLS)
                self.sig_changed = True
            self.frames = 0
        # Auto-Rotation at 30 fps to reduce workload
        if self.renderer.get_auto_rotate() and self.frames % 2 == 0:
            self.renderer.auto_rotate()
            self.sig_changed = True
        if self.sig_changed:
            try:
                self.renderer.buffer_scene()
                self.redraw()
                self.timeout = 0
            except curses.error:
                # Try again next update, this could hang, so do a timeout counter
                self.timeout += 1
        if self.timeout > 100:
            raise RuntimeError("Curses had an irrecoverable problem.")
        self.frames += 1
        return running

    def main_loop(self, main_screen):
        # The internal color setup requires curses to be initialized first
        conf.post_setup()
        # Save curses main screen for reference
        self.stdscr = main_screen
        # Init a new renderer of appropriate size
        self.renderer = Renderer(curses.LINES, curses.COLS)
        # Turns off cursor
        curses.curs_set(0)
        # Turns off hangup on input polling
        self.stdscr.nodelay(True)
        running = True
        self.redraw()
        while running:
            try:
                # Running at 60 fps
                sleep(1 / 60)
                running = self.on_update()
            except (KeyboardInterrupt, SystemError, SystemExit):
                running = False

    def run(self):
        if conf.parse():
            curses.wrapper(self.main_loop)
        return 0



import numpy as np

from asciimol.app.config import conf


class Renderer:
    def __init__(self, height, width):
        self.height = height
        self.width = width
        self.content = None
        self.zbuffer = None
        self.m = None
        self.f = 1.0
        self.resize(height, width)

        self.btoggle = len(conf.bonds) > 0
        self.pos = np.array(conf.coordinates)

        self.ztoggle = False
        self.zoom = 1.0
        self.rot = np.identity(3)
        self.rotcounter = [0, 0, 0]
        self.buffer_scene()

        self.auto_rotate_flags = np.array([False, False, False])

    def buffer_scene(self):
        """
        A super simple rasterizer. For now, just draw single character atom symbols at their rounded x and y
        positions.

        :return: True if nothing bad happened.
        """
        mx, my = self.m
        rot = np.matmul(self.pos, self.rot)
        self.clear()
        # Draw bonds
        for bond in conf.bonds:
            i, j = bond
            # if bond is (i, j) with i == j, just draw the label (no bonds)
            if i == j:
                x, y, z = rot[i]
                xp, yp = round(float(x) * self.f * self.zoom + mx), round(float(y) * self.zoom + my)
                if 1 < xp < self.width - 2 and 1 < yp < self.height - 3 and float(z) < self.zbuffer[yp][xp]:
                    self.zbuffer[yp][xp] = float(z)
                    self.content[yp][xp] = conf.symbols[0].upper() + "," + conf.colors[i]
            # else draw the bond with the labels at the end points
            else:
                # Draw the two labels at the end points
                xa, ya, za = rot[i]
                xa = float(xa) * self.f * self.zoom + mx
                ya = float(ya) * self.zoom + my
                xb, yb, zb = rot[j]
                xb = float(xb) * self.f * self.zoom + mx
                yb = float(yb) * self.zoom + my
                xap, yap = round(xa), round(ya)
                xbp, ybp = round(xb), round(yb)
                if 1 < xap < self.width - 2 and 1 < yap < self.height - 3 and float(za) < self.zbuffer[yap][xap]:
                    self.zbuffer[yap][xap] = float(za)
                    self.content[yap][xap] = conf.symbols[i].upper() + "," + conf.colors[i]
                if 1 < xbp < self.width - 2 and 1 < ybp < self.height - 3 and float(zb) < self.zbuffer[ybp][xbp]:
                    self.zbuffer[ybp][xbp] = float(zb)
                    self.content[ybp][xbp] = conf.symbols[j].upper() + "," + conf.colors[j]
                if not self.btoggle:
                    continue
                # Then start at xap+1 and go to xbp-1, drawing line segments
                sy = -1 if ya > yb else 1
                sx = -1 if xa > xb else 1
                sz = -1 if za > zb else 1
                dx = float((xb - xa) / (yb - ya)) if abs(yb - ya) > 0 else 0
                dy = float((yb - ya) / (xb - xa)) if abs(xb - xa) > 0 else 0
                dz = float((zb - za) / (xb - xa)) if abs(xb - xa) > 0 else 0
                if abs(dy) <= 1:
                    for k in range(1, abs(xap - xbp)):
                        xk = xap + sx * k
                        yk = round(float(ya) + sx * k * dy)
                        zk = round((float(za) + sz * k * dz))
                        if 1 < xk < self.width - 2 and 1 < yk < self.height - 3 and float(zk) < \
                                self.zbuffer[yk][xk]:
                            col = conf.colors[i] if k < abs(xap - xbp) / 2 else conf.colors[j]
                            self.zbuffer[yk][xk] = float(zk)
                            self.content[yk][xk] = "·,%s" % col
                else:
                    for k in range(1, abs(yap - ybp)):
                        xk = round((float(xa) + sy * k * dx))
                        yk = yap + sy * k
                        zk = round((float(za) + sz * k * dz))
                        if 1 < xk < self.width - 2 and 1 < yk < self.height - 3 and float(zk) < \
                                self.zbuffer[yk][xk]:
                            col = conf.colors[i] if k < abs(yap - ybp) / 2 else conf.colors[j]
                            self.zbuffer[yk][xk] = float(zk)
                            self.content[yk][xk] = "·,%s" % col
        return True

    def toggle_auto_rotate(self, x=False, y=False, z=False):
        self.auto_rotate_flags[[x, y, z]] = not all(self.auto_rotate_flags[[x, y, z]])

    def get_auto_rotate(self):
        return any(self.auto_rotate_flags)

    def auto_rotate(self):
        if self.auto_rotate_flags[0]:
            self.rotate(x=0.32)
        if self.auto_rotate_flags[1]:
            self.rotate(y=0.32)
        if self.auto_rotate_flags[2]:
            self.rotate(z=0.32)

    def rotate(self, x=0.0, y=0.0, z=0.0):
        """
        Set an internal rotation matrix that is applied to the coordinates before every render.
        """
        if abs(x) > 0:
            increment = np.pi / 36 * x
            sine = np.sin(increment)
            cosine = np.cos(increment)
            self.rot = np.matmul(self.rot, [[1.0, 0.0, 0.0], [0.0, cosine, -sine], [0.0, sine, cosine]])
            if self.rotcounter[0] + 5 * x > 360:
                self.rotcounter[0] -= 360
            elif self.rotcounter[0] + 5 * x < 0:
                self.rotcounter[0] += 360
            self.rotcounter[0] += 5 * x
        if abs(y) > 0:
            increment = np.pi / 36 * y
            sine = np.sin(increment)
            cosine = np.cos(increment)
            self.rot = np.matmul(self.rot, [[cosine, 0.0, sine], [0.0, 1.0, 0.0], [-sine, 0.0, cosine]])
            if self.rotcounter[1] + 5 * y > 360:
                self.rotcounter[1] -= 360
            elif self.rotcounter[1] + 5 * y < 0:
                self.rotcounter[1] += 360
            self.rotcounter[1] += 5 * y
        if abs(z) > 0:
            increment = np.pi / 36 * z
            sine = np.sin(increment)
            cosine = np.cos(increment)
            self.rot = np.matmul(self.rot, [[cosine, -sine, 0.0], [sine, cosine, 0.0], [0.0, 0.0, 1.0]])
            if self.rotcounter[2] + 5 * z > 360:
                self.rotcounter[2] -= 360
            elif self.rotcounter[2] + 5 * z < 0:
                self.rotcounter[2] += 360
            self.rotcounter[2] += 5 * z
        return abs(x) > 0 or abs(y) > 0 or abs(z) > 0

    def navigate(self, dx=0, dy=0):
        """
        Navigate the view so that the new mid point sits at (x + dx, y + dy).
        """
        x, y = self.m
        self.m = (x + dx, y + dy)
        return True

    def modify_zoom(self, factor):
        if self.zoom + factor > 0.2:
            self.zoom += factor
            return True
        return False

    def reset_view(self):
        """
        Reset the view to the starting values.
        """
        self.zoom = 1.0
        self.rotcounter = [0, 0, 0]
        self.rot = np.identity(3)
        self.m = round(self.width / 2), round(self.height / 2)
        self.pos = np.array(conf.coordinates)
        return True

    def resize(self, height, width):
        """
        Resize the screen. Known issue: crashes if the resize is faster than the framerate.
        """
        self.height = height
        self.width = width
        self.content = [[" ,0"] * self.width for _ in range(self.height - 2)]
        self.zbuffer = [[10000.0] * self.width for _ in range(self.height - 2)]
        self.m = round(self.width / 2), round(self.height / 2)
        # Since terminal characters are higher than wide, I correct for this by multiplying the x by f
        # so that it appears wider. 2.25 is what looks good on my terminals, but might be
        # nice to have a general way of determining the optimal value
        self.f = 2.25

    def clear(self):
        """
        Clear the canvas and redraw the border.
        """
        for i in range(self.height - 2):
            for j in range(self.width):
                self.zbuffer[i][j] = 10000.0
        for i in range(self.height - 2):
            for j in range(self.width):
                if i == 0 and j == 0:
                    self.content[i][j] = "┌,0"
                elif (i == 0 or i == self.height - 3) and 0 < j < self.width - 1:
                    self.content[i][j] = "─,0"
                elif i == 0 and j == self.width - 1:
                    self.content[i][j] = "┐,0"
                elif i < self.height - 3 and (j == 0 or j == self.width - 1):
                    self.content[i][j] = "│,0"
                elif i == self.height - 3 and j == 0:
                    self.content[i][j] = "└,0"
                elif i == self.height - 3 and j == self.width - 1:
                    self.content[i][j] = "┘,0"
                else:
                    self.content[i][j] = " ,0"

    def center(self):
        """
        Move the internal coordinate matrix to the center of coordinates
        """
        center = 1.0 / self.pos.shape[0] * np.sum(self.pos, axis=0)
        self.pos -= center
        return True

    def prinicple_axes(self):
        """
        Transform to principle axes of rotation
        """
        self.center()
        x, y, z = np.transpose(self.pos)
        xx = np.sum(y * y + z * z)
        yy = np.sum(x * x + z * z)
        zz = np.sum(x * x + y * y)
        xy = -np.sum(x * y)
        yz = -np.sum(y * z)
        xz = -np.sum(x * z)
        i = np.array([[xx, xy, xz], [xy, yy, yz], [xz, yz, zz]])
        w, t = np.linalg.eig(i)
        # Einstein Summation: for the 3x3 matrix t, multiply row j with each of the k rows in pos and sum
        # this is essentially just a basis transformation to the principle axes coordinate system.
        self.pos = np.einsum('ij,kj->ki', np.linalg.inv(t), self.pos)
        return True



import argparse
import tempfile

from asciimol.app import map_colors, map_radii
from asciimol.app.colors import init_curses_color_pairs


class Config:
    """
    The runtime configuration object. This is meant to be used as a singleton.
    Contains currently opened file, as well as any other configurables.
    """

    def __init__(self):
        self.coordinates = None
        self.symbols = None
        self.colors = None
        self.bonds = None

    def parse(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('XYZFILE', metavar='XYZFILE or SMILES', type=str,
                            help='Specify an .xyz file or a SMILES string (e.g., CC) to open and display.')
        opts = parser.parse_args()
        try:
            with open(opts.XYZFILE, "r") as handle:
                proceed, self.coordinates, self.symbols = read_xyz(handle)
        except ValueError:
            print("ERROR: Could not read '%s' as xyz file." % opts.XYZFILE)
            return False
        except FileNotFoundError:
            proceed, self.coordinates, self.symbols = read_smiles(opts.XYZFILE)
            if not proceed:
                print("ERROR: File '%s' not found and not a valid SMILES code!" % opts.XYZFILE)
                return False
        return proceed

    def post_setup(self):
        self._setup_bonds()
        self._setup_colors()

    def _setup_bonds(self):
        if self.bonds:
            return
        atms = len(self.symbols)
        radii = list(map_radii(self.symbols))
        bonds = []
        unbound = list(range(atms))
        for i in range(atms):
            for j in range(i):
                xa, ya, za = self.coordinates[i]
                xb, yb, zb = self.coordinates[j]
                rsq = (radii[i] + radii[j] + 0.41) ** 2
                dist = (xa - xb) ** 2 + (ya - yb) ** 2 + (za - zb) ** 2
                if dist < rsq or dist < 0.4:
                    bonds.append((i, j))
                    unbound[i] = -1
                    unbound[j] = -1
        for n, state in enumerate(unbound):
            if state != -1:
                bonds += [(n, n)]
        self.bonds = bonds

    def _setup_colors(self):
        if self.colors:
            return
        colors = list(map_colors(self.symbols))
        self.colors = list(init_curses_color_pairs(colors))


def read_smiles(smiles):
    from rdkit.Chem import AllChem, MolFromSmiles, AddHs
    from rdkit.Chem.rdmolfiles import MolToXYZBlock
    mol = MolFromSmiles(smiles)
    if mol is None:
        return False, None, None
    mol = AddHs(mol)
    AllChem.EmbedMolecule(mol)
    with tempfile.NamedTemporaryFile('w+') as temp:
        temp.writelines(MolToXYZBlock(mol))
        temp.seek(0)
        return read_xyz(temp)


def read_xyz(handle):
    line = handle.readline()
    try:
        atms = int(line.strip())
    except ValueError:
        print("XYZ FORMAT ERROR: Could not read atom number.")
        raise ValueError
    pos, sym = [], []
    handle.readline()  # Unused Comment line
    for n in range(atms):
        if line == "":
            print("XYZ FORMAT ERROR: Unexpected EOF. Atoms and Atom Number in line 1 mismatch!")
            raise ValueError
        line = handle.readline()
        work = line.strip().split()
        try:
            sym.append(work[0])
            pos.append([float(work[1]), float(work[2]), float(work[3])])
        except IndexError:
            print("XYZ FORMAT ERROR: Line '%s' is not formatted correctly." % line)
            raise ValueError
    return True, pos, sym


conf = Config()

__all__ = [conf]



import curses


def init_curses_color_pairs(a):
    if curses.has_colors():
        curses.use_default_colors()
        # Lazily init only colors as color pairs that are actually needed
        pair = 1
        cdict = dict()
        for entry in a:
            r, g, b, x256, x8 = entry
            col = x256 if curses.COLORS >= 256 else x8
            if str(col) not in cdict:
                try:
                    curses.init_pair(pair, col, -1)
                    cdict[str(col)] = pair
                    pair += 1
                except curses.error:
                    raise RuntimeError("Could not init pair %d with color %d" % (pair, col))
            yield str(curses.color_pair(cdict[str(col)]))
    else:
        return ["0"] * len(a)



from setuptools import setup, find_packages

with open('./README.md') as f:
    readme = f.read()

with open('./LICENSE') as f:
    lic = f.read()

packages = find_packages()

setup(
    name='asciimol',
    version='0.3.0',
    description='An ASCII molecule viewer.',
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Dominik Behrens',
    author_email='dewberryants@gmail.com',
    install_requires=['numpy', 'rdkit-pypi'],
    url='https://github.com/dewberryants/asciimol',
    license=lic,
    packages=find_packages(exclude="docs"),
    package_data={"": ["data/*"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Chemistry"
    ],
    entry_points={"console_scripts": ["asciimol = asciimol:run_asciimol"]}
)
